(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[5405],{75557:function(u,n,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/",function(){return t(56158)}])},56158:function(u,n,t){"use strict";t.r(n),t.d(n,{default:function(){return r}});var e=t(67294),s=t(11163);function r(){let u=(0,s.useRouter)();return(0,e.useEffect)(()=>{"/"===u.asPath?u.push("/clusters"):u.push(u.asPath)},[u]),null}}},function(u){u.O(0,[2888,9774,179],function(){return u(u.s=75557)}),_N_E=u.O()}]);