self.__BUILD_MANIFEST=function(s,c,a,e,t,b,f,u,n,o,j,i,r,k){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":["static/chunks/pages/index-444f1804401f04ea.js"],"/_error":["static/chunks/pages/_error-c66a4e8afc46f17b.js"],"/clusters":["static/chunks/pages/clusters-469814d711d63b1b.js"],"/clusters/[cluster]":[s,c,a,b,f,"static/chunks/4676-9da7fdbde90b5549.js",o,e,t,u,j,n,i,"static/chunks/6856-e6f350f567182e87.js",r,k,"static/chunks/9037-89a84fd7fa31362d.js","static/chunks/pages/clusters/[cluster]-ec747e4f2dc39b57.js"],"/clusters/[cluster]/[job]":[s,c,a,b,e,t,n,"static/chunks/pages/clusters/[cluster]/[job]-06afb50d25f7c61f.js"],"/config":["static/chunks/pages/config-dfb9bf07b13045f4.js"],"/infra":["static/chunks/pages/infra-c320641c2bcbbea6.js"],"/infra/[context]":["static/chunks/pages/infra/[context]-81351f95f3bec08e.js"],"/jobs":["static/chunks/pages/jobs-4b3ba1792dc6f21d.js"],"/jobs/pools/[pool]":[s,c,a,f,o,e,t,u,"static/chunks/pages/jobs/pools/[pool]-7d4182df6625fe10.js"],"/jobs/[job]":[s,c,a,b,f,o,e,t,u,n,"static/chunks/pages/jobs/[job]-ad2cd5aab787bc15.js"],"/users":["static/chunks/pages/users-018bf31cda52e11b.js"],"/volumes":["static/chunks/pages/volumes-739726d6b823f532.js"],"/workspace/new":["static/chunks/pages/workspace/new-3f88a1c7e86a3f86.js"],"/workspaces":["static/chunks/pages/workspaces-338de9df523d883a.js"],"/workspaces/[name]":[s,c,a,b,f,"static/chunks/6633-efe924b9b8136699.js",e,t,u,j,n,i,r,k,"static/chunks/1141-2f60a90b7d76838e.js","static/chunks/pages/workspaces/[name]-65f72dee417237ef.js"],sortedPages:["/","/_app","/_error","/clusters","/clusters/[cluster]","/clusters/[cluster]/[job]","/config","/infra","/infra/[context]","/jobs","/jobs/pools/[pool]","/jobs/[job]","/users","/volumes","/workspace/new","/workspaces","/workspaces/[name]"]}}("static/chunks/616-3d59f75e2ccf9321.js","static/chunks/6130-2be46d70a38f1e82.js","static/chunks/5739-d67458fcb1386c92.js","static/chunks/6989-01359c57e018caa4.js","static/chunks/3850-ff4a9a69d978632b.js","static/chunks/7411-b15471acd2cba716.js","static/chunks/1272-1ef0bf0237faccdb.js","static/chunks/8969-23c8fbdb8b397d59.js","static/chunks/6135-4b4d5e824b7f9d3c.js","static/chunks/754-d0da8ab45f9509e9.js","static/chunks/6990-08b2a1cae076a943.js","static/chunks/1121-2edb8ab2ba080a76.js","static/chunks/6601-06114c982db410b6.js","static/chunks/3015-bf218e4973bf5c8f.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();