(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[6324],{77614:function(e,n,r){(window.__NEXT_P=window.__NEXT_P||[]).push(["/workspaces/[name]",function(){return r(59431)}])},59431:function(e,n,r){"use strict";r.r(n),r.d(n,{default:function(){return i}});var u=r(85893);r(67294);var t=r(11163),l=r(11141);function i(){let e=(0,t.useRouter)(),{name:n}=e.query;return e.isReady&&n?(0,u.jsx)(l.WorkspaceEditor,{workspaceName:n}):(0,u.jsx)("div",{children:"Loading..."})}},9008:function(e,n,r){e.exports=r(37219)},12003:function(e,n,r){"use strict";r.d(n,{j:function(){return i}});var u=r(90512);let t=e=>"boolean"==typeof e?`${e}`:0===e?"0":e,l=u.W,i=(e,n)=>r=>{var u;if((null==n?void 0:n.variants)==null)return l(e,null==r?void 0:r.class,null==r?void 0:r.className);let{variants:i,defaultVariants:s}=n,o=Object.keys(i).map(e=>{let n=null==r?void 0:r[e],u=null==s?void 0:s[e];if(null===n)return null;let l=t(n)||t(u);return i[e][l]}),a=r&&Object.entries(r).reduce((e,n)=>{let[r,u]=n;return void 0===u||(e[r]=u),e},{});return l(e,o,null==n?void 0:null===(u=n.compoundVariants)||void 0===u?void 0:u.reduce((e,n)=>{let{class:r,className:u,...t}=n;return Object.entries(t).every(e=>{let[n,r]=e;return Array.isArray(r)?r.includes({...s,...a}[n]):({...s,...a})[n]===r})?[...e,r,u]:e},[]),null==r?void 0:r.class,null==r?void 0:r.className)}}},function(e){e.O(0,[616,6130,5739,7411,1272,6633,6989,3850,8969,6990,6135,1121,6601,3015,1141,2888,9774,179],function(){return e(e.s=77614)}),_N_E=e.O()}]);